class World {
  int coinsInStage; // how many coins are in the level

  int coinAnimDelay; // countdown timer between animation updates
  int coinAnimSpeed = 7; // speed between frame changes. Lower number = faster animation
  int coinAnimFrame; // keeps track of which animation frame is currently shown for the player
  
  static final int TILE_EMPTY = 0;
  static final int TILE_SOLID = 1;
  static final int TILE_COIN = 2;
  static final int TILE_KILLBLOCK = 3;
  static final int TILE_START = 4;
  static final int TILE_EARTH = 5;
  static final int TILE_CLOUD = 6;
  static final int TILE_CURVEL = 7;
  static final int TILE_CURVER = 8;
  static final int TILE_EDGEL = 9;
  static final int TILE_EDGER = 10;
  static final int TILE_BOTTOM = 11;
  static final int TILE_SIDEL = 12;
  static final int TILE_SIDER = 13;
  static final int TILE_WALLTHIN = 14;
  static final int TILE_SPIKE = 15;
  static final int TILE_SOLIDTHIN = 16;
  static final int TILE_GRASS = 17;
  
  static final int GRID_UNIT_SIZE = 48; // size, in pixels, of each world unit square
  // if the above number becomes too small, how the player's wall bumping is detected may need to be updated

  // these next 2 numbers need to match the dimensions of the example level grid below
  static final int GRID_UNITS_WIDE =120;
  static final int GRID_UNITS_TALL = 13;

  int[][] worldGrid = new int[GRID_UNITS_TALL][GRID_UNITS_WIDE]; // the game checks this one during play
  
  // the game copies this into worldGrid each reset, returning coins that have since been cleared
  int[][] start_Grid = { { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13, 5, 5, 5, 5,12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13, 5, 5, 5, 5,12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13, 5, 5, 5, 5,12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13, 5, 5, 5, 5,12,17, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,17,17, 0, 0, 0, 0,17, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 1, 1, 1, 1, 1, 1, 1, 9,10, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,17, 0, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 1, 1, 1, 1, 1, 1, 1, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,16, 0, 0, 0, 8, 5, 5, 5, 5, 5, 5, 7, 0, 2,10, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 9, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 1, 1, 1, 1, 5, 5, 5, 5,11, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,17, 0, 0,14, 0,17, 0, 0,13, 5, 5,11, 5,12, 0, 0, 0, 2,10, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 9, 2, 0, 0, 0, 0, 0},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,17, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,17, 0, 0, 0, 8,11, 5, 5, 5, 5, 5, 5, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 1, 1, 1, 1, 1, 9, 0,13, 5,12, 2,13,12, 0, 0, 0, 0, 2,10, 1, 9, 0, 0, 0, 0, 0, 0,10, 1, 9, 2, 0, 0, 0, 0, 0, 0},
                         { 0, 4, 0,17, 0, 0, 0, 0, 0, 0, 0, 0,17, 0, 0, 0,10, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0,10, 1, 1, 1, 9, 0, 0, 8, 5, 5, 5, 5,12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,17, 0, 0, 0, 0, 0, 0,17, 0, 0, 0,17, 0, 0, 0, 0, 0,16, 0, 0, 0, 0, 0, 8, 5,11,11, 5, 7, 0, 0,13, 5,12, 2,13,12, 0, 0, 0, 0, 0, 2,10, 1, 9, 0, 0, 0, 0,10, 1, 9, 2, 0, 0, 0, 0, 0, 0, 0},
                         {10, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 9, 0, 0, 8, 1, 1, 9, 0, 0, 0,16, 0, 0, 0, 0, 8, 5, 7, 0, 0, 0, 0,13, 5, 5, 5,12, 0, 0, 0, 0, 0, 0,16, 0, 0, 0, 0, 0, 0, 0,10, 1, 1, 1, 1, 1, 9, 0, 0, 0,10, 1, 1, 1, 1, 9, 0, 0, 0,16, 0,14, 0, 0, 0,16, 0, 0,14, 2, 2,14, 0, 0, 0,13, 5,12, 0,13,12, 0, 0, 0, 0, 0, 0, 2,10, 1, 9, 0, 0,10, 1, 9, 2, 0, 0, 0, 0, 0, 0, 0, 0},
                         { 2, 8, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,11, 7, 0, 0, 0, 0,13,12, 0, 0, 0, 0,14, 0, 0, 0, 0, 2,14, 2, 0, 0, 0,10, 1, 1, 1, 1, 1, 9, 0, 0,17, 0, 0,14, 0, 0, 0,16, 0, 0, 0, 0, 8, 5, 5, 5,12, 0, 0, 0, 0, 0,13, 5, 5, 7, 0, 0, 0, 0,14, 0,14, 0, 0, 0,14, 0, 0,14, 2, 2,14, 0, 0, 0,13, 5,12, 0,13,12, 0, 0, 0, 0, 0, 0, 0, 2,10, 1, 9,10, 1, 9, 2, 0, 0, 0, 2, 0, 0, 0, 0,17},
                         {15,15, 8, 5, 5, 5, 5, 5, 5, 5, 5,12,15,15,15,15,15,15,13,12,15,15,15,15,14,15,15,15,15,15,14,15,15,15,15,15, 8, 5, 1, 1, 1, 1, 1, 1, 1, 9,15,14,15,15,15,14,15,15,15,15,15, 8, 5, 5,12,15,15,15,15,15,13, 5, 7,15,15,15,15,15,14,15,14,15,15,15,14,15,15,14,15,15,14,15,15,15,13, 5,12,15,13,12,15,15,15,15,15,15,15,15,15,14,15,15,14,15,15,15,15,15,15,15,15,15,15,10},
                         { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} };
  
  // returns what type of tile is at a given pixel coordinate
  int worldSquareAt(PVector thisPosition) {
    float gridSpotX = thisPosition.x/GRID_UNIT_SIZE;
    float gridSpotY = thisPosition.y/GRID_UNIT_SIZE;
  
    // first a boundary check, to avoid looking outside the grid
    // if check goes out of bounds, treat it as a solid tile (wall) - Could be killblock etc.
    if(gridSpotX<0) {
      return TILE_SOLID; 
    }
    if(gridSpotX>=GRID_UNITS_WIDE) {
      return TILE_SOLID; 
    }
    if(gridSpotY<0) {
      return TILE_SOLID; 
    }
    if(gridSpotY>=GRID_UNITS_TALL) {
      return TILE_KILLBLOCK;
    }
    
    return worldGrid[int(gridSpotY)][int(gridSpotX)];
  }
  
  // changes the tile at a given pixel coordinate to be a new tile type
  // currently used to replace TILE_COIN tiles with TILE_EMPTY tiles once collected
  void setSquareAtToThis(PVector thisPosition, int newTile) {
    int gridSpotX = int(thisPosition.x/GRID_UNIT_SIZE);
    int gridSpotY = int(thisPosition.y/GRID_UNIT_SIZE);
  
    if(gridSpotX<0 || gridSpotX>=GRID_UNITS_WIDE || 
       gridSpotY<0 || gridSpotY>=GRID_UNITS_TALL) {
      return; // can't change grid units outside the grid
    }
    
    worldGrid[gridSpotY][gridSpotX] = newTile;
  }
  
  // these helper functions help us correct for the player moving into a world tile
  float topOfSquare(PVector thisPosition) {
    int thisY = int(thisPosition.y);
    thisY /= GRID_UNIT_SIZE;
    return float(thisY*GRID_UNIT_SIZE);
  }
  float bottomOfSquare(PVector thisPosition) {
    if(thisPosition.y<0) {
      return 0;
    }
    return topOfSquare(thisPosition)+GRID_UNIT_SIZE;
  }
  float leftOfSquare(PVector thisPosition) {
    int thisX = int(thisPosition.x);
    thisX /= GRID_UNIT_SIZE;
    return float(thisX*GRID_UNIT_SIZE);
  }
  float rightOfSquare(PVector thisPosition) {
    if(thisPosition.x<0) {
      return 0;
    }
    return leftOfSquare(thisPosition)+GRID_UNIT_SIZE;
  }
  
  void reload() {
    coinsInStage = 0; // we count them while copying in level data
    
    for(int i=0;i<GRID_UNITS_WIDE;i++) {
      for(int ii=0;ii<GRID_UNITS_TALL;ii++) {
        if(start_Grid[ii][i] == TILE_START) { // player start position
          worldGrid[ii][i] = TILE_EMPTY; // put an empty tile in that spot
  
          // then update the player spot to the center of that tile
          thePlayer.position.x = i*GRID_UNIT_SIZE+(GRID_UNIT_SIZE/2);
          thePlayer.position.y = ii*GRID_UNIT_SIZE+(GRID_UNIT_SIZE);
        } else {
          if(start_Grid[ii][i]==TILE_COIN) {
            coinsInStage++;
          }
          worldGrid[ii][i] = start_Grid[ii][i];
        }
      }
    }
  }
  
  void render() { // draw the world
    
     //animation logic for coins
       if(coinAnimDelay--<0) {
        coinAnimDelay=coinAnimSpeed;
        if(coinAnimFrame==0) {
          coinAnimFrame=1;
        } else if(coinAnimFrame==1){
          coinAnimFrame=2;
        } else if(coinAnimFrame==2){
          coinAnimFrame=3;
        } else {
          coinAnimFrame=0;
        }
      }
    
    //assign graphics to tiles - could be images, could be rectangles, etc 
    for(int i=0;i<GRID_UNITS_WIDE;i++) { // for each column
      for(int ii=0;ii<GRID_UNITS_TALL;ii++) { // for each tile in that column
        
        if(worldGrid[ii][i]==TILE_SOLID) {
          image(ground,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_SPIKE) {
          image(killBlock,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_EARTH) {
          image(earth,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_GRASS) {
          image(grass,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_CURVEL) {
          image(curveLeft,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_CURVER) {
          image(curveRight,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_CLOUD) {
          image(cloud,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_EDGEL) {
          image(groundEdgeL,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_EDGER) {
          image(groundEdgeR,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_BOTTOM) {
          image(bottomEarth,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_SIDEL) {
          image(sideL,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_SIDER) {
          image(sideR,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_WALLTHIN) {
          image(wallThin,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }
        if(worldGrid[ii][i]==TILE_SOLIDTHIN) {
          image(solidThin,i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // x,y of top left corner to draw rectangle
             GRID_UNIT_SIZE,GRID_UNIT_SIZE);
        }        
        if(worldGrid[ii][i]==TILE_COIN) { // if it's a coin, draw the coin
          image(coin[coinAnimFrame],i*GRID_UNIT_SIZE,ii*GRID_UNIT_SIZE, // center of grid spot
                  GRID_UNIT_SIZE,GRID_UNIT_SIZE); // spin size wide, 1/2 height of box tall
        }
      }
    }
  }
}
